/*
 *  Version: 1.0
 *
 *  The contents of this file are subject to the OpenVPMS License Version
 *  1.0 (the 'License'); you may not use this file except in compliance with
 *  the License. You may obtain a copy of the License at
 *  http://www.openvpms.org/license/
 *
 *  Software distributed under the License is distributed on an 'AS IS' basis,
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 *  for the specific language governing rights and limitations under the
 *  License.
 *
 *  Copyright 2008 (C) OpenVPMS Ltd. All Rights Reserved.
 *
 *  $Id$
 */

package org.openvpms.web.component.im.query;

import org.openvpms.component.business.domain.im.party.Party;
import org.openvpms.component.system.common.query.ArchetypeQuery;
import org.openvpms.component.system.common.query.Constraints;
import org.openvpms.component.system.common.query.ObjectSelectConstraint;
import org.openvpms.component.system.common.query.ObjectSet;
import org.openvpms.component.system.common.query.ShortNameConstraint;
import org.openvpms.component.system.common.query.SortConstraint;

import java.util.Date;


/**
 * An {@link org.openvpms.web.component.im.query.ResultSet} implementation that queries patients. The search can be
 * further constrained to match on:
 * <ul>
 * <li>partial patient name; and/or
 * <li>partial identity
 * </ul>
 * <p/>
 * The returned {@link org.openvpms.component.system.common.query.ObjectSet}s contain:
 * <ul>
 * <li>the patient:
 * <pre>Party patient = (Party) set.get("patient");</pre>
 * <li>the identity, if searching on identities:
 * <pre>EntityIdentity identity = (EntityIdentity) set.get("identity");</pre>
 * </ul>
 *
 * @author <a href="mailto:support@openvpms.org">OpenVPMS Team</a>
 * @version $LastChangedDate: 2006-05-02 05:16:31Z $
 */
public class PatientResultSet extends AbstractEntityResultSet<ObjectSet> {

    /**
     * The customer to return patients for. If <tt>null</tt>, queries all patients.
     */
    private Party customer;

    /**
     * Creates a new <tt>PatientResultSet</tt>.
     *
     * @param archetypes       the archetypes to query
     * @param value            the value to query on. May be <tt>null</tt>
     * @param searchIdentities if <tt>true</tt> search on identity name
     * @param customer         if specified, only return patients for the specified customer
     * @param sort             the sort criteria. May be <tt>null</tt>
     * @param rows             the maximum no. of rows per page
     */
    public PatientResultSet(ShortNameConstraint archetypes, String value, boolean searchIdentities,
                            Party customer, SortConstraint[] sort, int rows) {
        super(archetypes, value, searchIdentities, null, sort,
              rows, true, new ObjectSetQueryExecutor());
        archetypes.setAlias("patient");
        this.customer = customer;
    }

    /**
     * Determines if all patients are being queried.
     *
     * @return <tt>true</tt> if all patients are being queried, or <tt>false</tt> if only the patients for the
     *         specified customer are being returned.
     */
    public boolean isSearchingAllPatients() {
        return customer == null;
    }

    /**
     * Creates a new archetype query.
     *
     * @return a new archetype query
     */
    @Override
    protected ArchetypeQuery createQuery() {
        ArchetypeQuery query = super.createQuery();
        query.add(new ObjectSelectConstraint("patient"));

        if (customer != null) {
            Date now = new Date();
            query.add(Constraints.join("customers", Constraints.shortName("rel", "entityRelationship.patientOwner")));
            query.add(Constraints.eq("rel.source", customer.getObjectReference()));
            query.add(Constraints.lte("rel.activeStartTime", now));
            query.add(Constraints.or(Constraints.gte("rel.activeEndTime", now),
                                     Constraints.isNull("rel.activeEndTime")));
        }

        if (isSearchingIdentities()) {
            query.add(new ObjectSelectConstraint("identity"));
        }
        return query;
    }

}